from typing import Optional, Dict, List
from ragmatic.utils.refs import RefBaseModel
from pydantic import Field

from ...actions._types import (
    DocumentSourceComponentConfig,
    StorageComponentConfig,
    LLMComponentConfig,
    SummarizerComponentConfig,
    EncoderComponentConfig,
    RagAgentComponentConfig,
)
from ...actions.bases import ActionConfig
from ragmatic.common_types import TypeAndConfig


class RagQueryCommandConfig(RefBaseModel):
    rag_agent: RagAgentComponentConfig
    document_source: TypeAndConfig


class ComponentConfig(RefBaseModel):
    document_sources: Optional[Dict[str, DocumentSourceComponentConfig]] = Field(default=None)
    storage: Optional[Dict[str, StorageComponentConfig]] = Field(default=None)
    llms: Optional[Dict[str, LLMComponentConfig]] = Field(default=None)
    summarizers: Optional[Dict[str, SummarizerComponentConfig]] = Field(default=None)
    encoders: Optional[Dict[str, EncoderComponentConfig]] = Field(default=None)
    rag_agents: Optional[Dict[str, RagAgentComponentConfig]] = Field(default=None)


class PipelineElementConfig(RefBaseModel):
    action: str
    config: ActionConfig


class MasterConfig(RefBaseModel):
    project_name: Optional[str] = Field(default=None) 
    components: Optional[ComponentConfig] = Field(default=None)
    pipelines: Optional[Dict[str, List[PipelineElementConfig]]] = Field(default=None)
    rag_query_command: Optional[RagQueryCommandConfig] = Field(default=None)



from . import bases


class PyCodeRagAgent(bases.RagAgentBase):

    name = "python_code"
    system_prompt = (
        "You are an assistant with expertise in programming, particularly with "
        "python applications. You pay careful attention to detail "
        "and provide clear, concise answers that avoid hyperbole."
    )
    prompt = (
        "I'm a python developer, and I have a question about some code in a "
        "codebase. I'll first ask my question, and then I'll share a number of "
        "code files all from the codebase. Please answer my question as it "
        "relates to these files. Here's the question, followed by a divider and "
        "then the codebase files:"
    )

    file_filters = [(lambda x: x.endswith('.py'))]
    
    def build_user_message(self, query: str, context_docs: dict[str, str]):
        context_block = ""
        for relpath, doc in context_docs.items():
            context_block += f"```python file={relpath}\n{doc}\n```\n\n"
        return (
            self.prompt +
            "\n" +
            query +
            self.q_context_delimiter +
            context_block
        )
    


import typing as t
from ragmatic.utils.refs import RefBaseModel
from pydantic import ConfigDict
from ragmatic.common_types import TypeAndConfig

class ActionConfig(RefBaseModel):

    document_source: t.Optional[TypeAndConfig] = None
    model_config = ConfigDict(extra= "allow")


class Action:
    
    config_cls: ActionConfig = None
    name: str = None

    def __init__(self, config):
        self.config = self._config_from_action_config(config)
        
    def execute(self):
        raise NotImplementedError

    def _config_from_action_config(self, config: ActionConfig):
        return self.config_cls(**config.model_dump())



from .bases import Action
from .encode import EncodeAction
from .summarize import SummarizeAction


_actions = {
    EncodeAction.name: EncodeAction,
    SummarizeAction.name: SummarizeAction,
}


def get_action_cls(action_name: str) -> Action:
    if action_name not in _actions:
        raise ValueError(f"Action '{action_name}' not found.")
    return _actions[action_name]


def list_actions() -> list[str]:
    return sorted(list(_actions.keys()))



import typing as t
from logging import getLogger

from ragmatic.utils.refs import RefBaseModel

from ragmatic.storage.store_factory import get_store_cls
from ragmatic.storage.bases import TextDocumentStore, VectorStore
from ragmatic.embeddings.embedder_factory import get_embedder_cls, Embedder
from ._types import EncoderComponentConfig, StorageComponentConfig
from .bases import Action, ActionConfig
from ..document_sources.source_factory import get_document_source_cls, DocumentSourceBase
from ragmatic.common_types import TypeAndConfig


logger = getLogger(__name__)


class EncodeActionConfig(ActionConfig):
    encoder: EncoderComponentConfig
    document_source: TypeAndConfig
    storage: StorageComponentConfig
    

class EncodeAction(Action):

    config_cls = EncodeActionConfig
    name = "encode"

    def __init__(self, config: EncodeActionConfig):
        super().__init__(config)
        self._embedder: Embedder = self._initialize_encoder()
        self._source: DocumentSourceBase = self._initialize_source()
        self._vector_store: VectorStore = self._initialize_storage()

    def _initialize_encoder(self):
        embedder_cls = get_embedder_cls(self.config.encoder.type)
        embedder_config = self.config.encoder.config
        return embedder_cls(embedder_config)
    
    def _initialize_storage(self, data_type=None, type_=None, config=None) -> TextDocumentStore:
        data_type = data_type or self.config.storage.data_type
        type_ = type_ or self.config.storage.type
        store_cls = get_store_cls(data_type, type_)
        store_config = config or self.config.storage.config
        return store_cls(store_config)
    
    def _initialize_source(self) -> DocumentSourceBase:
        source_cls = get_document_source_cls(self.config.document_source.type)
        source_config = self.config.document_source.config
        return source_cls(source_config)
    
    def execute(self):
        logger.info("Loading source data")
        source_data = self._source.get_documents()
        logger.info(f"Encoding {len(source_data)} documents...")
        _encoded_data = self._embedder.encode([
            doc for _, doc in source_data.items()
        ])
        embeddings = {
            key: value for key, value in zip(source_data.keys(), _encoded_data)
        }
        logger.info(f"Storing {len(embeddings)} embeddings...")
        self._vector_store.store_vectors(embeddings)
        logger.info(f"Embeddings stored in {self._vector_store.name} storage.")



from ragmatic.utils.refs import Ref
from ._types import PresetData
from .local_docs_preset import local_docs_preset


variable_defaults = local_docs_preset.variable_defaults.copy()
variable_defaults.pop("local_documents_path")
variable_defaults["local_python_package_path"] = "./src"

_component_config = local_docs_preset.components.copy()
_component_config.update({
    "document_sources": {
        "local_python_package": {
            "type": "pycode_filesystem",
            "config": {
                "root_path": "${local_python_package_path}"
            }
        },
        "localpy_storage": {
            "type": "storage",
            "config": Ref("components.storage.localpy")
        }
    },
    "rag_agents": {
        "pycode": {
            "type": "python_code",
            "config": {
                "llm": Ref("components.llms.openai"),
                "storage": Ref("components.storage.localpy"),
                "encoder": Ref("components.encoders.plaintext"),
                "n_nearest": "${n_nearest}",
                "prompt": "",
                "system_prompt": ""
            }
        }
    }
})


_pipelines_config = {
    "ingest-python-codebase": [
        {
            "action": "summarize",
            "config": {
                "document_source": Ref("components.document_sources.local_python_package"),
                "root_path": "./src",
                "summarizer": Ref("components.summarizers.pycode"),
                "storage": Ref("components.storage.localpy")
            }
        },
        {
            "action": "encode",
            "config": {
                "document_source": Ref("components.document_sources.local_python_package"),
                "encoder": Ref("components.encoders.plaintext"),
                "storage": Ref("components.storage.localpy")
            }
        }
    ]
}


_rag_query_command_config = local_docs_preset.rag_query_command.copy()
_rag_query_command_config.update({
    "rag_agent": Ref("components.rag_agents.pycode"),
    "document_source": Ref("components.document_sources.local_python_package")
})


pycode_preset = PresetData(
    components=_component_config,
    pipelines=_pipelines_config,
    rag_query_command=_rag_query_command_config,
    variable_defaults=variable_defaults
)



from .bases import SummarizerBase
from .py_code_summarizer import PyCodeSummarizer
from ragmatic.utils import import_object

_summarizers = {
    PyCodeSummarizer.name: PyCodeSummarizer
}


def get_summarizer_class(summarizer_name: str) -> SummarizerBase:
    if summarizer_name not in _summarizers:
        try:
            return import_object(summarizer_name)
        except Exception:
            raise ValueError(f"Summarizer {summarizer_name} not found")
    return _summarizers[summarizer_name]



from anthropic import Anthropic

from .bases import LLMClientBase, ContentBase, LLMState


class AnthropicContent(ContentBase):

    def get_content(self):
        return {
            "type": "text",
            "text": self.msg
        }


class AnthropicClient(LLMClientBase):

    content_type = AnthropicContent

    def __init__(self, config: dict):
        super().__init__(config)
        self.model = config.get("model", "claude-3-5-sonnet-20240620")
        self.default_headers = config.get("default_headers", {})
        self.max_tokens = config.get("max_tokens", 4096)
        self.client = self._get_client()

    def _get_client(self):
        return Anthropic(
            api_key=self._plaintextkey(),
            default_headers=self.default_headers
        )

    def _load_api_key(self) -> str:
        key = super()._load_api_key()
        if not key:
            raise ValueError("anthropic API key not found.")
        return key

    def send_message(self,
                     message: str,
                     system_prompt: str = None,
                     role: str = "user",
                     ) -> str:
        messages = [{
            "role": role,
            "content": AnthropicContent(message).get_content()
        }]
        
        response = self.client.messages.create(
            model=self.model,
            messages=messages,
            max_tokens=self.max_tokens,
            system_prompt=[system_prompt] if system_prompt else []
        )
        content = response.content
        return "\n\n".join([c.text for c in content])

    def send_chat(self, state: LLMState) -> str:
        messages = state.messages

        final_role = messages[-1]["role"]
        if final_role != "user":
            raise ValueError("The final message must be from the user.")

        response = self.client.messages.create(
            model=state.model,
            messages=messages,
            max_tokens=self.max_tokens,
            system_prompt=[state.system_prompt] if state.system_prompt else []
        )
        content = response.content
        return "\n\n".join([c.text for c in content])




import logging
import os
import re
import typing as t
from itertools import product
from logging import getLogger

import numpy as np
import torch
import torch.nn.functional as F
import tqdm
from ragmatic.utils.refs import RefBaseModel
from pydantic import ConfigDict
from sklearn.preprocessing import normalize
from transformers import AutoModel, AutoModelForCausalLM, AutoTokenizer

from .bases import Embedder

logger = getLogger(__name__)


def _get_salesforce_model_names():
    def _get_model_name(size, data):
        return f"Salesforce/codegen-{size}-{data}"

    sizes = ["350M", "2B", "6B", "16B"]
    data = ["nl", "mono", "multi"]
    return [_get_model_name(size, data) for size, data in product(sizes, data)]


class HfTransformersEmbeddingConfig(RefBaseModel):
    model_config = ConfigDict(protected_namespaces=())
    model_name: str
    tokenizer_config: dict = {}
    save_filepath: str = "embedding_model.pkl"
    save_model: bool = True
    expected_hidden_size: int = 1024
    chunk_size: int = 512
    overlap: int = 128

class HfTransformersEmbedder(Embedder):
    embedder_name = "hf_transformer"
    _causal_lm_models = {
        *_get_salesforce_model_names(),
    }

    def __init__(self, config):
        self.config = config
        self.model_name = config["model_name"]
        self.tokenizer_config = config.get("tokenizer_config", {})
        self.save_model: bool = config.get("save_model", True)
        self.save_filepath = config.get("save_filepath", "embedding_model.pkl")
        self._expected_hidden_size = config.get("expected_hidden_size", 1024)
        self._observed_hidden_size = 0
        self._auto_model_class = self._init_auto_model_class()
        self._model = None
        self._tokenizer = None

    @property
    def model(self):
        if not self._model:
            if os.path.exists(self.save_filepath):
                self._load_model()
            else:
                self._download_model()
        assert self._model is not None
        return self._model

    @property
    def tokenizer(self):
        _ = self.model
        assert self._tokenizer is not None
        return self._tokenizer

    def _silence_loggers(self):
        to_silence = ["^transformers", "^torch"]
        prefix_re = re.compile(rf'^(?:{ "|".join(to_silence) })')
        for name in logging.root.manager.loggerDict:
            if re.match(prefix_re, name):
                logging.getLogger(name).setLevel(logging.ERROR)

    def _download_model(self):
        logger.info(f"Loading model {self.model_name!r}")
        self._model = self._auto_model_class.from_pretrained(self.model_name, trust_remote_code=True)
        if self.save_model:
            logger.info(f"Saving model to {self.save_filepath}")
            self._model.save(self.save_filepath)
        logger.info(f"Loading tokenizer {self.model_name!r}")
        self._tokenizer = AutoTokenizer.from_pretrained(self.model_name, trust_remote_code=True)
        if self.save_model:
            tokenizer_filepath = (
                os.path.dirname(self.save_filepath) + "/tokenizer.pkl"
            )
            logger.info(f"Saving model to {tokenizer_filepath}")
            self._tokenizer.save_pretrained(tokenizer_filepath)

    def _load_model(self):
        logger.info(f"Loading model from {self.save_filepath}")
        self._model = self._auto_model_class(
            self.save_filepath, trust_remote_code=True
        )
        tokenizer_filepath = (
            os.path.dirname(self.save_filepath) + "/tokenizer.pkl"
        )
        logger.info(f"Loading tokenizer from  {tokenizer_filepath}")
        self._tokenizer = AutoTokenizer.from_pretrained(tokenizer_filepath)

    def encode(self, docs: t.Sequence[str], query: bool = False) -> t.Sequence[t.Sequence[float]]:
        embeddings = []
        for doc in tqdm.tqdm(docs):
            if not doc:
                zero_length = (self._observed_hidden_size * 2) or (
                    self._expected_hidden_size * 2
                )
                embeddings.append(np.zeros((1, zero_length)))
                continue
            embeddings.append(self._encode_doc(doc))
        return embeddings

    def _init_auto_model_class(self):
        if self.model_name in self._causal_lm_models:
            return AutoModelForCausalLM
        return AutoModel

    def chunk_text(self, text, chunk_size=None, overlap=None):
        chunk_size = chunk_size or self.chunk_size
        overlap = overlap or self.overlap
        tokens = self.tokenizer.tokenize(text)
        chunks = []
        for i in range(0, len(tokens), chunk_size - overlap):
            chunk = tokens[i : i + chunk_size]
            chunks.append(self.tokenizer.convert_tokens_to_string(chunk))
        return chunks

    def _encode_doc(self, doc):
        chunks = self.chunk_text(doc)
        embeddings = []
        for chunk in chunks:
            embedding = self._encode_chunk(chunk)
            if isinstance(embedding, np.ndarray):
                embedding = torch.from_numpy(embedding)
            embeddings.append(embedding)
        embedding_stack = torch.stack(embeddings)
        return torch.mean(embedding_stack, dim=0).numpy()

    def _encode_chunk(self, doc):
        if not self.tokenizer.pad_token and self.tokenizer.eos_token:
            self.tokenizer.pad_token = self.tokenizer.eos_token
        else:
            self.tokenizer.pad_token = '[PAD]'
        inputs = self.tokenizer(doc, **self.tokenizer_config)
        with torch.no_grad():
            outputs = self.model(**inputs, output_hidden_states=True)
        last_hidden_state = outputs.hidden_states[-1]
        self._observed_hidden_size = last_hidden_state.shape[-1]
        last_state_mean_pooled = self.process_hidden_state(
            last_hidden_state, inputs["attention_mask"]
        )
        next_last_hidden_state = outputs.hidden_states[-3]
        next_last_state_mean_pooled = self.process_hidden_state(
            next_last_hidden_state,
            inputs["attention_mask"],
            pooling_strategy="attention",
        )
        return normalize(
            torch.cat(
                [last_state_mean_pooled, next_last_state_mean_pooled], dim=-1
            )
        )

    def process_hidden_state(
        self, hidden_state, attention_mask, pooling_strategy="mean"
    ):
        # assumes hidden_state.size is (batch_size, seq_len, hidden_size)
        real_tokens_mask = (
            attention_mask.unsqueeze(-1).expand(hidden_state.size()).float()
        )
        masked_hidden_state = hidden_state * real_tokens_mask
        if pooling_strategy == "mean":
            sum_embeddings = torch.sum(masked_hidden_state, dim=1)
            sum_mask = torch.clamp(
                attention_mask.sum(dim=1, keepdim=True), min=1e-9
            )
            mean_pooled = sum_embeddings / sum_mask
            return mean_pooled
        elif pooling_strategy == "max":
            max_pooled, _ = torch.max(masked_hidden_state, dim=1)
            return max_pooled
        elif pooling_strategy == "attention":
            cls_vector = hidden_state[:, 0, :]
            attention_scores = torch.matmul(
                hidden_state, cls_vector.unsqueeze(-1)
            ).squeeze(-1)
            attention_scores = F.softmax(attention_scores, dim=-1)
            attention_pooled = torch.bmm(
                attention_scores.unsqueeze(1), hidden_state
            ).squeeze(1)
            return attention_pooled
        else:
            raise ValueError(f"Invalid pooling strategy {pooling_strategy}")



class CollectionKeyFormatter:

    delim = "::"

    @staticmethod
    def flatten_collection_key(collection_name, index):
        return f"{collection_name}{CollectionKeyFormatter.delim}{index}"

    @staticmethod
    def extract_collection_name(key):
        return key.split(CollectionKeyFormatter.delim)[0]



import typing as t
from pydantic import Field
from logging import getLogger

from ragmatic.utils import CollectionKeyFormatter
from ._types import SummarizerComponentConfig, StorageComponentConfig
from ..storage.bases import TextDocumentStore
from ..storage.store_factory import get_store_cls
from ..summarization.bases import SummarizerBase
from ..summarization.summarizer_factory import get_summarizer_class
from ..document_sources.source_factory import get_document_source_cls, DocumentSourceBase
from .bases import Action, ActionConfig
from ragmatic.common_types import TypeAndConfig

logger = getLogger(__name__)


class SummarizeActionConfig(ActionConfig):
    summarizer: SummarizerComponentConfig
    storage: StorageComponentConfig
    document_source: TypeAndConfig


class SummarizeAction(Action):

    config_cls = SummarizeActionConfig
    name = "summarize"
    
    def __init__(self, config: SummarizeActionConfig):
        super().__init__(config)
        self._text_doc_store: TextDocumentStore = self._initialize_storage()
        self._summarizer: SummarizerBase = self._initialize_summarizer()
        self._source: DocumentSourceBase = self._initialize_document_source()
        
    def _initialize_summarizer(self):
        summarizer_cls = get_summarizer_class(self.config.summarizer.type)
        summarizer_config = self.config.summarizer.config
        return summarizer_cls(summarizer_config)

    def _initialize_document_source(self):
        source_cls = get_document_source_cls(self.config.document_source.type)
        source_config = self.config.document_source.config
        return source_cls(source_config)

    def _initialize_storage(self) -> TextDocumentStore:
        store_cls = get_store_cls(
            self.config.storage.data_type,
            self.config.storage.type
        )
        store_config = self.config.storage.config
        return store_cls(store_config)

    def execute(self):
        documents = self._source.get_documents()
        summaries = self._summarizer.summarize(documents)
        kv_summaries = self.summaries_to_key_value_pairs(summaries)
        logger.info(f"Summarization completed. {len(summaries)} docs summarized with {len(kv_summaries)} summaries.")
        self._text_doc_store.store_text_docs(kv_summaries)
        logger.info(f"Summaries stored in {self._text_doc_store.name} storage.")


    def summaries_to_key_value_pairs(self, summaries: dict[str, list[str]]) -> dict[str, str]:
        return {
            CollectionKeyFormatter.flatten_collection_key(doc_name, i): summary
            for doc_name, summary_texts in summaries.items()
            for i, summary in enumerate(summary_texts)
        }


import logging
import os

logging.basicConfig(
    level=os.environ.get("RAGMATIC_LOG_LEVEL", "INFO"),
    format="{message}",
    style="{"
)


import click
from .commands import (
    rag_cmd,
    run_cmd
)


@click.group()
def cli():
    pass


cli.add_command(rag_cmd)
cli.add_command(run_cmd)



import typing as t
from ragmatic.utils.refs import RefBaseModel


class TypeAndConfig(RefBaseModel):
    type: str
    config: dict


class StoreConfig(RefBaseModel):
    data_type: str
    type: str
    config: dict



from typing import Type
from ragmatic.utils import import_object
from .bases import Embedder


_embedders = {
    "hf_transformers": "ragmatic.embeddings.hugging_face_transformer.HfTransformersEmbedder",
    "hf_sentence_transformers": "ragmatic.embeddings.hf_sentence_transformers.HfSentenceTransformersEmbedder",
}


def get_embedder_cls(embedder_name) -> Type[Embedder]:
    if embedder_name not in _embedders:
        try:
            return import_object(embedder_name)
        except Exception:
            raise ValueError(f"Embedder {embedder_name} not supported")
    return import_object(_embedders[embedder_name])



from ._types import PresetData
from ragmatic.utils.refs import Ref

variable_defaults = {
    "local_documents_path": "./documents",
    "encoding_model_name": "dunzhang/stella_en_400M_v5",
    "expected_hidden_size": 1024,
    "n_nearest": 10
}

doc_sources = {
    "local_directory": {
        "type": "filesystem",
        "config": {
            "root_path": "${local_documents_path}"
        }
    }
}

storage = {
    "localpy": {
        "data_type": "omni",
        "type": "pydict",
        "config": {
            "dirpath": "./data",
            "overwrite": True
        }
    }
}

llms = {
    "openai": {
        "type": "openai",
        "config": {
            "api_keyenvvar": "OPENAI_API_KEY",
        }
    }
}

summarizers = {
    "pycode": {
        "type": "python_code",
        "config": {
            "document_source": doc_sources["local_directory"],
            "llm": llms["openai"],
        }
    }
}

encoders = {
    "plaintext": {
        "type": "hf_sentence_transformers",
        "config": {
            "model_name": "${encoding_model_name}",
            "query_prompt_name": "s2p_query",
            "chunk_size": 512,
            "overlap": 128
        }
    },
    "plaintext_transformers": {
        "type": "hf_transformers",
        "config": {
            "model_name": "${encoding_model_name}",
            "tokenizer_config": {
                "return_tensors": "pt",
                "max_length": 512,
                "truncation": True,
                "padding": "longest"
            },
            "save_model": False,
            "expected_hidden_size": "${expected_hidden_size}",
            "chunk_size": 512,
            "overlap": 128
        }
    }
}

rag_agents = {
    "generic": {
        "type": "generic",
        "config": {
            "llm": Ref("components.llms.openai"), 
            "storage": Ref("components.storage.localpy"),
            "encoder": Ref("components.encoders.plaintext"), 
            "n_nearest": "${n_nearest}",
            "prompt": "",
            "system_prompt": ""
        }
    }
}

_component_config = {
    "document_sources": doc_sources,
    "storage": storage,
    "llms": llms,
    "summarizers": summarizers,
    "encoders": encoders,
    "rag_agents": rag_agents
}

_pipelines_config = {
    "ingest-directory-text": [
        {
            "action": "encode",
            "config": {
                "document_source": Ref("components.document_sources.local_directory"),
                "encoder": Ref("components.encoders.plaintext"),
                "storage": Ref("components.storage.localpy")
            }
        }
    ]
}

_rag_query_command_config = {
    "rag_agent": Ref("components.rag_agents.generic"),
    "document_source": Ref("components.document_sources.local_directory")
}


local_docs_preset = PresetData(
    components=_component_config,
    pipelines=_pipelines_config,
    rag_query_command=_rag_query_command_config,
    variable_defaults=variable_defaults
)



from typing import List, Callable
from logging import getLogger

from ragmatic.utils.refs import RefBaseModel
from tqdm.contrib.concurrent import thread_map

from ..llm_ops.bases import LLMClientBase
from ..llm_ops.client_factory import get_llm_client_class
from ragmatic.common_types import TypeAndConfig


logger = getLogger(__name__)


class SummarizerConfig(RefBaseModel):
    llm: TypeAndConfig
    document_source: TypeAndConfig


class SummarizerBase:
    summarizer_name: str = ""
    _document_prompt: str = ""
    _system_prompt: str = ""
    file_filters: List[Callable[[str], bool]] = [(lambda x: True)]

    def __init__(self, config: dict):
        self.config: SummarizerConfig = config
        self._llm_client: LLMClientBase = self._initialize_llm_client()
        self._summaries: dict[str, list[str]] = {}
    
    def _initialize_llm_client(self) -> LLMClientBase:
        client_class = get_llm_client_class(self.config.llm.type)
        llm_config = self.config.llm.config
        return client_class(llm_config)

    def summarize(self, documents: dict[str, str] = None) -> dict[str, list[str]]:
        docs = documents
        _jobs = [(name, text) for name, text in docs.items()]
        logger.info(f"Summarizing {len(_jobs)} code documents...")
        summary_responses = thread_map(self.summarize_document, [doc for _, doc in _jobs])
        self._summaries = dict(zip([doc_name for doc_name, _ in _jobs], summary_responses))
        return self._summaries

    def summarize_document(self, doc: str) -> list[str]:
        """
        Example implementation:

        ```python
        message = self._build_message(doc)
        response = self._llm_client.send_message(
            message,
            self._system_prompt,
            role="user"
        )
        return response.split("\n")
        ```
        """
        raise NotImplementedError
    
    def _build_message(self, code_doc: str) -> str:
        return self._document_prompt + "\n---\n" + code_doc
    
    def _file_path_to_doc_name(self, file_path: str) -> str:
        return file_path
    


import typing as t
from pydantic import Field
from logging import getLogger

from ragmatic.utils import CollectionKeyFormatter
from ..storage.bases import TextDocumentStore
from ..storage.store_factory import get_store_cls
from ..summarization.bases import SummarizerBase
from ..summarization.summarizer_factory import get_summarizer_class
from .bases import Action, ActionConfig
from ..rag.bases import RagAgentConfig, RagAgentBase
from ._types import RagAgentComponentConfig
from ..rag.rag_agent_factory import get_rag_agent_class
from ..document_sources.source_factory import get_document_source_cls
from ..document_sources.bases import DocumentSourceBase
from ragmatic.common_types import TypeAndConfig


logger = getLogger(__name__)


class RagActionConfig(ActionConfig):
    rag_agent: t.Union[str, RagAgentComponentConfig]
    document_source: t.Union[str, TypeAndConfig]
    query: str


class RagAction(Action):

    config_cls = RagActionConfig
    name = "rag"

    def __init__(self, config: RagActionConfig):
        super().__init__(config)
        self._source: DocumentSourceBase = self._initialize_source()
        self._rag_agent: RagAgentBase = self._initialize_rag_agent()

    def _initialize_source(self):
        source_cls = get_document_source_cls(self.config.document_source.type)
        source_config = self.config.document_source.config
        return source_cls(source_config)

    def _initialize_rag_agent(self):
        rag_agent_cls = get_rag_agent_class(self.config.rag_agent.type)
        return rag_agent_cls(self.config.rag_agent.config, self._source)
    
    def execute(self) -> dict[str, str]:
        return self._rag_agent.query(self.config.query)



from logging import getLogger
from typing import Union, get_args, get_origin
from pydantic import BaseModel, ConfigDict
from pydantic._internal._model_construction import ModelMetaclass

import yaml
import json
import copy


logger = getLogger(__name__)


class Ref:
    def __init__(self, loc):
        self.loc = loc

    def __eq__(self, value: object) -> bool:
        return isinstance(value, Ref) and value.loc == self.loc


def ref_constructor(loader, node):
    value = loader.construct_scalar(node)
    return Ref(value)


def ref_representer(dumper, data):
    return dumper.represent_scalar(u'!ref', data.loc)


yaml.add_constructor(u'!ref', ref_constructor, yaml.SafeLoader)
yaml.add_representer(Ref, ref_representer)


def resolve_references(data, root=None):
    if root is None:
        root = data

    if isinstance(data, dict):
        return {key: resolve_references(value, root) for key, value in data.items()}
    elif isinstance(data, list):
        return [resolve_references(item, root) for item in data]
    elif isinstance(data, Ref):
        resolved = resolve_ref(root, data.loc)
        return resolve_references(resolved, root)
    else:
        return data

def resolve_ref(data, loc):
    parts = loc.split('.')
    current = data
    for part in parts:
        if part in current:
            current = current[part]
        else:
            raise KeyError(f"Reference not found: {loc}")
    return copy.deepcopy(current)


def ragmatic_load_yaml(stream):
    return yaml.safe_load(stream)


def ref_dumper_default(obj):
    if isinstance(obj, Ref):
        return f"!ref {obj.loc}"
    raise TypeError(f"Object of type {type(obj)} is not JSON serializable")


class RefDecoder(json.JSONDecoder):
    
    def __init__(self, *args, **kwargs):
        super().__init__(object_hook=self._object_hook, *args, **kwargs)

    def _object_hook(self, obj):
        if isinstance(obj, dict):
            return {key: self._object_hook(value) for key, value in obj.items()}
        elif isinstance(obj, list):
            return [self._object_hook(item) for item in obj]
        if isinstance(obj, str) and obj.startswith("!ref "):
            obj: str = obj
            loc = obj.replace('!ref ', '')
            return Ref(loc)
        return obj


def make_refable(typ):
    return Union[typ, Ref]


class RefModelMetaclass(ModelMetaclass):

    def __new__(mcs, name, bases, namespace, **kwargs):
        annotations = namespace.get('__annotations__', {})
        for field_name, field_type in annotations.items():
            if field_name.startswith('__'):
                continue
            annotations[field_name] = make_refable(field_type)
        namespace['__annotations__'] = annotations
        return super().__new__(mcs, name, bases, namespace, **kwargs)


class RefBaseModel(BaseModel, metaclass=RefModelMetaclass):

    @classmethod
    def model_validate(cls, obj, *args, **kwargs):
        if isinstance(obj, dict):
            for field_name, field_value in obj.items():
                if isinstance(field_value, Ref):
                    continue
                field_info = cls.model_fields[field_name]
                field_type = get_args(field_info.annotation)[0]  # Get the non-Ref type
                if get_origin(field_type) is not None:
                    field_type = get_origin(field_type)
                if isinstance(field_type, type) and issubclass(field_type, RefBaseModel):
                    obj[field_name] = field_type.model_validate(field_value)
        return super().model_validate(obj, *args, **kwargs)

    model_config = ConfigDict(arbitrary_types_allowed=True)


import typing as t
from ragmatic.utils import import_object


_vector_stores = {
    "pydict": "ragmatic.storage.pydict.vector_store.PydictVectorStore"
}

_text_doc_stores =  {
    "pydict": "ragmatic.storage.pydict.text_doc_store.PydictTextDocumentStore"
}

_omni_stores = {
    "pydict": "ragmatic.storage.pydict.omni_store.PydictOmniStore"
}


def get_store_cls(data_type:str, store_type: str) -> t.Type:
    store_dict = globals()[f"_{data_type}_stores"]
    if store_type not in store_dict:
        try:
            return import_object(store_type)
        except Exception:
            raise ValueError(f"Store {store_type} not found")
    return import_object(store_dict[store_type])



import logging
import re
import typing as t
from logging import getLogger

import numpy as np
import torch
import tqdm
from ragmatic.utils.refs import RefBaseModel
from pydantic import ConfigDict, Field
from sentence_transformers import SentenceTransformer

from .bases import Embedder

logger = getLogger(__name__)



class HfSentenceTransformersEmbeddingConfig(RefBaseModel):
    model_config = ConfigDict(protected_namespaces=())
    model_name: str
    query_prompt_name: str = Field(default="s2p_query")
    chunk_size: int = Field(default=512)
    overlap: int = Field(default=128)

class HfSentenceTransformersEmbedder(Embedder):
    embedder_name = "hf_sentence_transformer"

    def __init__(self, config):
        self.config = config
        self.model_name = config["model_name"]
        self.query_prompt_name = config["query_prompt_name"]
        self.chunk_size = config.get("chunk_size", 512)
        self.overlap = config.get("overlap", 128)
        self._observed_hidden_size = 0
        self._model = None

    @property
    def model(self):
        if not self._model:
            self._download_model()
        assert self._model is not None
        return self._model


    def _silence_loggers(self):
        to_silence = ["^transformers", "^torch", "^sentence_transformers"]
        prefix_re = re.compile(rf'^(?:{ "|".join(to_silence) })')
        for name in logging.root.manager.loggerDict:
            if re.match(prefix_re, name):
                logging.getLogger(name).setLevel(logging.ERROR)

    def _download_model(self):
        logger.info(f"Loading model {self.model_name!r}")
        self._model = SentenceTransformer(self.model_name, trust_remote_code=True)
        
    def _load_model(self):
        self._download_model()

    def encode(self, docs: t.Sequence[str], query: bool = False) -> t.Sequence[t.Sequence[float]]:
        embeddings = []
        for doc in tqdm.tqdm(docs):
            if not doc:
                zero_length = self._observed_hidden_size
                embeddings.append(np.zeros((1, zero_length)))
                continue
            embeddings.append(self._encode_doc(doc))
        return embeddings

    def chunk_text(self, text: str, chunk_size=None, overlap=None):
        chunk_size = chunk_size or self.chunk_size
        overlap = overlap or self.overlap
        tokens = text.split()
        chunks = []
        longest = 0
        for i in range(0, len(tokens), chunk_size - overlap):
            chunk = tokens[i : i + chunk_size]
            longest = max(longest, len(chunk))
            chunks.append(chunk)
        if len(chunks) > 1:
            for i in range(len(chunks)):
                chunk = chunks[i]
                if len(chunk) < longest:
                    new_bits = ["[PAD]"] * (longest - len(chunk))
                    chunk.extend(new_bits)
                    chunks[i] = " ".join(chunk)
        return [" ".join(c) for c in chunks]

    def _encode_doc(self, doc, query=False):
        query_prompt = self.query_prompt_name if query else None
        chunks = self.chunk_text(doc)
        embeddings = []
        for chunk in chunks:
            embedding = self._encode_chunk(chunk, query_prompt)
            if isinstance(embedding, np.ndarray):
                embedding = torch.from_numpy(embedding)
            embeddings.append(embedding)
        embedding_stack = torch.stack(embeddings)
        return torch.mean(embedding_stack, dim=0).numpy()

    def _encode_chunk(self, doc, query_prompt=None):
        if query_prompt:
            return self.model.encode(doc, prompt_name=query_prompt)
        return self.model.encode(doc)



from openai import OpenAI
import logging

from .bases import LLMClientBase, ContentBase, LLMState


class OpenAIContent(ContentBase):

    def get_content(self):
        return self.msg


class OpenAIClient(LLMClientBase):

    content_type = OpenAIContent

    def __init__(self, config: dict):
        super().__init__(config)
        self.model = config.get("model", "gpt-4o")
        self._log_level = config.get("log_level", logging.WARNING)
        self.client = self._get_client()
        self._set_openai_log_level(self._log_level)

    def _set_openai_log_level(self, level: str):
        logging.getLogger("openai").setLevel(level)
        logging.getLogger("httpx").setLevel(level)

    def _get_client(self):
        return OpenAI(api_key=self._plaintextkey())

    def _load_api_key(self) -> str:
        key = super()._load_api_key()
        if not key:
            raise ValueError("openai API key not found.")
        return key

    def send_message(self,
                     message: str,
                     system_prompt: str = None,
                     role: str = "user",
                     ) -> str:
        messages = []
        if system_prompt:
            messages.append({"role": "system", "content": system_prompt})
        messages.append({"role": role, "content": OpenAIContent(message).get_content()})
        response = self.client.chat.completions.create(
            model=self.model,
            messages=messages    
        )
        return response.choices[0].message.content.strip()
    
    def send_chat(self, state: LLMState) -> str:
        messages = state.messages

        final_role = messages[-1]["role"]
        if final_role != "user":
            raise ValueError("The final message must be from the user.")

        if state.system_prompt:
            penultimate_role = messages[-2]["role"]
            if penultimate_role == "system":
                messages[-2]["content"] = state.system_prompt
            else:
                messages.insert(-1, {"role": "system", "content": state.system_prompt})
        
        response = self.client.chat.completions.create(
            model=state.model,
            messages=messages    
        )
        return response.choices[0].message.content.strip()




from abc import ABC, abstractmethod
import typing as t


class Embedder:

    embedder_name: str = None

    @abstractmethod
    def __init__(self, config):
        pass

    @abstractmethod
    def encode(self, docs: t.Sequence[str], query: bool = False) -> t.Sequence[t.Sequence[float]]:
        pass



import click
from logging import getLogger

from ragmatic.cli.configuration.tools import (
    load_config,
    merge_defaults,
    MasterConfig,
    resolve_config_references
)
from ragmatic.actions.rag import RagActionConfig, RagAction
from ..configuration.presets.preset_factory import get_preset


logger = getLogger(__name__)


@click.command('rag-query')
@click.option('--query', type=str, required=True)
@click.option('--config', type=click.Path(exists=True))
@click.option('--preset', type=click.Choice(['local_docs', 'pycode']), default='local_docs')
@click.option('--var', '-v', type=str, multiple=True, help="Variables to override in the preset, in key=var format")
def rag_cmd(config: click.Path, query: str, preset: str, var: list):
    vars = dict(v.split('=') for v in var)
    preset_data = get_preset(preset)
    preset_config = preset_data.get_config(**vars)
    config: MasterConfig = load_config(config) if config else preset_config
    if config != preset_config:
        config = merge_defaults(config, preset_data, **vars)
    config = resolve_config_references(config)
    cmd_config = config.rag_query_command
    rag_agent_component_config = cmd_config.rag_agent
    rag_action_config = RagActionConfig(
        **dict(
            document_source=cmd_config.document_source,
            query=query,
            rag_agent=rag_agent_component_config.model_dump()
        )
    )
    print(RagAction(rag_action_config).execute())



from .factory import import_object
from .collection_key_formatter import CollectionKeyFormatter
from .refs import ragmatic_load_yaml


ALLOWED_FILE_TYPES = [
    'babelconfig', 'babelrc', 'bash', 'c', 'cc', 'cfg', 'clj', 'cljc', 'cljs', 'conf',
    'cpp', 'css', 'csv', 'dart', 'dockerignore', 'docx', 'editorconfig', 'edn',
    'env', 'eslintrc', 'fish', 'flowconfig', 'gitattributes', 'gitignore', 'go',
    'graphqlconfig', 'groovy', 'h', 'hpp', 'html', 'ini', 'ipynb', 'java',
    'jestconfig', 'js', 'json', 'jsx', 'kt', 'kts', 'less', 'lock', 'log', 'md',
    'odf', 'odg', 'odp', 'ods', 'odt', 'otf', 'otg', 'otp', 'ots', 'ott', 'pdf',
    'php', 'pl', 'pm', 'postcssconfig', 'pptx', 'prettierrc', 'properties', 'py',
    'r', 'rb', 'rmd', 'rs', 'rst', 'sass', 'scala', 'scss', 'sh', 'sql',
    'stylelintrc', 'swift', 't', 'toml', 'ts', 'tsconfig', 'tsv', 'tsx', 'txt',
    'webpackconfig', 'xlsx', 'xml', 'yaml', 'yml', 'zsh'
]



import typing as t


class DocumentSourceBase:

    name: str = ""

    def __init__(self, config):
        self.config = config

    def get_documents(self, document_names: t.Optional[list[str]] = None) -> dict[str, str]:
        raise NotImplementedError




import typing as t
import pickle
import os
import numpy as np
import re
from collections import OrderedDict
from logging import getLogger

from ragmatic.utils.refs import RefBaseModel
from pydantic import Field

from ..bases import VectorStore


logger = getLogger(__name__)


class QueryMethod:
    name: str = None

    @staticmethod
    def execute(query: dict, data: dict[str, np.ndarray]):
        pass

    @staticmethod
    def build_query_from_embedding(embedding: t.Sequence[float]) -> t.Any:
        pass


class CosineSimilarity(QueryMethod):
    """
    Expected query format:
    {
        "method": "cosine_similarity",
        "vector": <np.ndarray>,
        "limit": <int> (optional)
    }
    """

    name = 'cosine_similarity'

    @staticmethod
    def execute(query: dict, data: dict[str, np.ndarray]):
        if 'vector' not in query:
            raise ValueError('Query must contain a "vector" key value pair')
        query_vector = query['vector']
        if not isinstance(query_vector, np.ndarray):
            raise ValueError('Query vector must be a numpy array')
        data = OrderedDict(data)
        doc_embeddings_matrix = np.asarray(list(data.values()))
        logger.debug(f"Matrix shape: {doc_embeddings_matrix.shape}")
        query_vector, doc_embeddings_matrix =\
            CosineSimilarity._check_and_reshape(
                query_vector,
                doc_embeddings_matrix
            )
        similarities =\
            CosineSimilarity._cosine_similarity(
                query_vector,
                doc_embeddings_matrix
            ).flatten()
        sorted_indices_desc = np.argsort(similarities)[::-1].flatten()
        results = [list(data.keys())[i] for i in sorted_indices_desc]
        if limit := query.get('limit'):
            return results[:limit]
        return results
    
    @staticmethod
    def _cosine_similarity(v1_single: np.ndarray, matrix: np.ndarray):
        v1 = v1_single.flatten()
        dot_product = np.dot(matrix, v1)
        norm_v1 = np.linalg.norm(v1)
        norm_matrix = np.linalg.norm(matrix, axis=1)
        norm_product = norm_v1 * norm_matrix
        cosine_similarity = dot_product / norm_product
        return cosine_similarity

    @staticmethod
    def _check_and_reshape(vector, matrix):
        if vector.ndim == 1:
            logger.info("Reshaping vector to (1, n)")
            vector = vector.reshape(1, -1)
        
        if vector.ndim != 2 or vector.shape[0] != 1:
            raise ValueError(f"Vector should have shape (1, n), but has shape {vector.shape}")
        
        if matrix.ndim not in [2, 3]:
            raise ValueError(f"Matrix should be 2D or 3D, but has {matrix.ndim} dimensions")
        
        if matrix.ndim == 3:
            if matrix.shape[1] != 1:
                raise ValueError(f"For 3D matrix, second dimension should be 1, but shape is {matrix.shape}")
            matrix = matrix.reshape(matrix.shape[0], matrix.shape[2])
        
        if vector.shape[1] != matrix.shape[1]:
            raise ValueError(f"Vector dimension ({vector.shape[1]}) does not match matrix last dimension ({matrix.shape[1]})")
        
        return vector, matrix


class PydictVectorStoreConfig(RefBaseModel):
    filepath: t.Optional[str] = Field(default='vectors.pkl')
    default_query_method: t.Optional[str] = Field(default="cosine_similarity")
    overwrite: t.Optional[bool] = Field(default=False)
    allow_init: t.Optional[bool] = Field(default=True)


class PydictVectorStore(VectorStore):
    
    name = 'pydict'
    _allowed_query_methods = {
        "cosine_similarity": CosineSimilarity
    }

    def __init__(self, config):
        config = PydictVectorStoreConfig(**config)
        self.config = config
        self.overwrite = config.overwrite
        self.allow_init = config.allow_init
        self.filepath = os.path.expanduser(self.config.filepath)
        self.__data: dict[str, np.ndarray] = {}
        self._default_query_method = config.default_query_method

    @property
    def _data(self):
        if not self.__data:
            self._load_vectors()
        return self.__data

    def store_vectors(self, vectors: dict[str, np.ndarray]):
        if self.overwrite:
            self.__data = vectors
        else:
            self._data.update(vectors)
        logger.info(f"Storing vectors to {self.filepath}")
        self._write_vectors(self._data)

    def get_vectors(self, keys: list[str]):
        return [self._data.get(key) for key in keys]

    def scan_keys(self, match: str):
        return [key for key in self._data.keys() if re.match(match, key)]

    def _write_vectors(self, data):
        with open(self.filepath, "wb") as f:
            pickle.dump(data, f)

    def _load_vectors(self):
        if not os.path.exists(self.filepath):
            if self.allow_init:
                return
            raise FileNotFoundError(
                f"Vector data not loaded: File {self.filepath} does not exist."
            )
        with open(self.filepath, "rb") as f:
            self.__data = pickle.load(f)
    
    def query(self, query: dict):
        if 'method' not in query:
            query["method"] = self._default_query_method
        method = query['method']
        if method not in self._allowed_query_methods:
            raise ValueError(f'Invalid query method: {method}')
        return self._allowed_query_methods[method].execute(query, self._data)
    
    def query_byvector(self, vector: t.Sequence[float], n: int = None):
        return self.query({
            "vector": vector,
            "limit": n
        })



from logging import getLogger
import typing as t

import click

from ..configuration._types import MasterConfig
from ..configuration.tools import (
    load_config,
    merge_defaults,
    resolve_config_references
)
from ...actions.bases import Action
from ...actions.action_factory import get_action_cls
from ..configuration.presets.preset_factory import get_preset

logger = getLogger(__name__)


@click.command('run-pipeline')
@click.argument('pipeline', type=str, required=False, default=None)
@click.option('--config', type=click.Path(exists=True))
@click.option('--preset', type=click.Choice(['local_docs', 'pycode']), default='local_docs')
@click.option('--var', '-v', type=str, multiple=True, help="Variables to override in the preset, in key=var format")
def run_cmd(pipeline: t.Union[str, None], config: click.Path, preset: str, var: list):
    vars = dict(v.split('=') for v in var)
    preset_data = get_preset(preset)
    preset_config = preset_data.get_config(**vars)
    config: MasterConfig = load_config(config) if config else preset_config
    if config != preset_config:
        config = merge_defaults(config, preset_data, **vars)
    config = resolve_config_references(config)
    if not config.pipelines:
        raise ValueError(f"No pipelines found in config: {config}")
    if pipeline is None:
        first_preset_pipeline_name = list(preset_config.pipelines.keys())[0]
        if first_preset_pipeline_name not in config.pipelines:
            pipeline = list(config.pipelines.keys())[0]
        else:
            pipeline = first_preset_pipeline_name
    pipeline_config = config.pipelines[pipeline]
    logger.info(f"Running pipeline: {pipeline}")
    for element in pipeline_config:
        action_name = element.action
        action_cls: t.Type[Action] = get_action_cls(action_name)
        action: Action = action_cls(element.config)
        logger.info(f"Running action: {action_name}")
        action.execute()



from .bases import DocumentSourceBase
from .filesystem import FilesystemDocumentSource, PycodeFilesystemDocumentSource
from .storage import TextStoreDocumentSource
from ragmatic.utils import import_object


_sources = {
    FilesystemDocumentSource.name: FilesystemDocumentSource,
    PycodeFilesystemDocumentSource.name: PycodeFilesystemDocumentSource,
    TextStoreDocumentSource.name: TextStoreDocumentSource,
}


def get_document_source_cls(name: str) -> DocumentSourceBase:
    if name not in _sources:
        try:
            return import_object(name)
        except Exception:
            raise ValueError(f"Document source '{name}' not found.")
    return _sources[name]



from .bases import RagAgentBase
from ragmatic.utils import ALLOWED_FILE_TYPES

class GenericRagAgent(RagAgentBase):

    name = "generic"
    system_prompt = (
        "You are an assistant with attention to detail who is able to provide "
        "clear answers to questions relating to a collection of documents."
    )
    prompt = (
        "I have a question about some documents in a collection. I'll first ask "
        "my question, and then I'll share a number of documents, all from the collection. "
        "Please answer my question as it relates to these documents, and include "
        "any relevant sections of text from the documents in your response. "
        "Here's the question, followed by a divider and then the documents:"
    )

    file_filters = [(lambda x: x.split(".")[-1] in ALLOWED_FILE_TYPES)]
    
    def build_user_message(self, query: str, context_docs: dict[str, str]):
        context_block = ""
        for relpath, doc in context_docs.items():
            context_block += f"```txt file={relpath}\n{doc}\n```\n\n"
        return (
            self.prompt +
            "\n" +
            query +
            self.q_context_delimiter +
            context_block
        )
    


from typing import Type
from ragmatic.utils import import_object
from .bases import LLMClient
from ragmatic.utils import import_object

_clients = {
    "openai": "ragmatic.llm_ops.openai_.OpenAIClient",
    "anthropic": "ragmatic.llm_ops.anthropic_.AnthropicClient"
}


def get_llm_client_class(client_name: str) -> Type[LLMClient]:
    if client_name not in _clients:
        try:
            return import_object(client_name)
        except Exception:
            raise ValueError(f"Client name {client_name} not found in available clients: {list(_clients)}")
    return import_object(_clients[client_name])



import os
import typing as t

from ragmatic.utils.refs import RefBaseModel
from pydantic import Field

from ..bases import OmniStore
from .vector_store import PydictVectorStore
from .text_doc_store import PydictTextDocumentStore


class PydictOmniStoreConfig(RefBaseModel):
    dirpath: t.Optional[str] = Field(default=None)
    overwrite: t.Optional[bool] = Field(default=True)


class PydictOmniStore(OmniStore):
    
    name = "pydict"
    _default_dirpath = 'data'

    def __init__(self, config):
        config = PydictOmniStoreConfig(**config)
        self.config = config
        self.overwrite = config.overwrite
        self.dirpath = os.path.expanduser((config.dirpath or self._default_dirpath))
        os.makedirs(self.dirpath, exist_ok=True)
        self._vector_store = self._init_vector_store()
        self._text_doc_store = self._init_text_doc_store()

    def _init_vector_store(self):
        filepath = os.path.join(self.dirpath, 'vectors.pkl')
        config = self._substore_config(filepath)
        return PydictVectorStore(config)
    
    def _init_text_doc_store(self):
        filepath = os.path.join(self.dirpath, 'text_documents.pkl')
        config = self._substore_config(filepath)
        return PydictTextDocumentStore(config)

    def _substore_config(self, filepath):
        return {
            "filepath": filepath,
            "overwrite": self.overwrite
        }



import os
from typing import Any, Optional
from dataclasses import dataclass, field
import base64
from abc import ABC, abstractmethod



@dataclass
class MessageBox:
    msg: str


class ContentBase:

    def __init__(self, msg: str):
        self.msg = msg

    def get_content(self):
        raise NotImplementedError

@dataclass
class LLMState:
    model: str
    client: Any
    content_type: ContentBase
    messages: list[dict] = field(default_factory=list)
    system_prompt: Optional[str] = None


class LLMClient(ABC):

    content_type: ContentBase = None

    @abstractmethod
    def __init__(self, config: dict):
        pass

    @abstractmethod
    def send_message(self,
                     message: str,
                     system_prompt: str = None,
                     role: str = "user",
                     ) -> str:
        pass

    @abstractmethod
    def send_chat(self, state: LLMState) -> str:
        pass


class LLMClientBase(LLMClient):

    def __init__(self, config: dict):
        self.config = config
        self._api_keyenvvar = config.get("api_keyenvvar", "")
        self._api_keypath = config.get("api_keypath", "")
        self._api_key = self._load_api_key()

    def _load_api_key(self) -> str:
        if self._api_keyenvvar:
            _key = os.environ.get(self._api_keyenvvar)
            if not _key:
                return None
            return self._b64key(_key)
        if self._api_keypath:
            with open(self._api_keypath, "r") as f:
                return self._b64key(f.read().strip())
        
    def _b64key(self, keystring: str) -> str:
        return base64.b64encode(keystring.encode()).decode()

    def _plaintextkey(self) -> str:
        return base64.b64decode(self._api_key.encode()).decode()



import typing as t
from ragmatic.utils.refs import RefBaseModel
from pydantic import Field, ConfigDict

from ..summarization.bases import SummarizerConfig

from ..rag.bases import RagAgentConfig


class DocumentSourceComponentConfig(RefBaseModel):
    type: t.Literal["storage", "filesystem", "pycode_filesystem"]
    config: t.Union[str, dict] = Field(default_factory=dict)


class LLMComponentConfig(RefBaseModel):
    type: str
    config: dict
    model_config = ConfigDict(extra= "allow")


class EncoderComponentConfig(RefBaseModel):
    type: str
    config: dict = Field(default_factory=dict)


class AnalysisConfig(RefBaseModel):
    analyzer_type: t.Literal["python"]
    storage: str


class SummarizerComponentConfig(RefBaseModel):
    type: t.Literal["python_code"]
    config: SummarizerConfig


class StorageComponentConfig(RefBaseModel):
    data_type: t.Literal["metadata", "vector", "summary", "omni"]
    type: t.Literal["pydict"]
    config: dict = Field(default_factory=dict)


class RagAgentComponentConfig(RefBaseModel):
    type: str
    config: RagAgentConfig



import os
import typing as t
from ragmatic.utils.refs import RefBaseModel
from .bases import DocumentSourceBase
from ragmatic.utils import ALLOWED_FILE_TYPES


class FilesystemDocumentSourceConfig(RefBaseModel):
    root_path: str


class FilesystemDocumentSource(DocumentSourceBase):
    
    name = "filesystem"
    file_filters: t.List[t.Callable[[str], bool]] = [(lambda x: x.split(".")[-1] in ALLOWED_FILE_TYPES)]

    def __init__(self, config: FilesystemDocumentSourceConfig):
        config = FilesystemDocumentSourceConfig(**config)
        super().__init__(config)
        self.root_path = os.path.abspath(config.root_path)
        if not os.path.exists(self.root_path):
            raise FileNotFoundError(f"Root path {self.root_path} does not exist.")

    def get_documents(self, document_names: t.Optional[list[str]] = None) -> dict[str, str]:
        if document_names is None:
            return self._get_all_documents()
        return self._get_documents_by_names(document_names)
    
    def _get_documents_by_names(self, document_names: list[str]) -> dict[str, str]:
        documents = {}
        for name in document_names:
            path = self._document_name_to_file_path(name)
            with open(path, "r") as f:
                documents[name] = f.read()
        return documents
    
    def _get_all_documents(self) -> dict[str, str]:
        walked = list(os.walk(self.root_path))
        documents = {}
        for root, _, files in walked:
            for file in files:
                if all([f(file) for f in self.file_filters]):
                    file_path = os.path.join(root, file)
                    doc_name = self._file_path_to_doc_name(file_path)
                    with open(file_path, 'r') as file:
                        doc = file.read()
                    documents[doc_name] = doc
        return documents

    def _document_name_to_file_path(self, doc_name: str) -> str:
        return os.path.join(self.root_path, doc_name)

    def _file_path_to_doc_name(self, file_path: str) -> str:
        return os.path.relpath(file_path, self.root_path)


class PycodeFilesystemDocumentSource(FilesystemDocumentSource):

    name = "pycode_filesystem"
    file_filters: t.List = [(lambda x: x.endswith('.py'))]
    
    def _document_name_to_file_path(self, doc_name: str) -> str:
        if not doc_name.endswith('.py'):
            # assume a module name
            doc_name = doc_name.replace('.', os.path.sep) + '.py'
        return super()._document_name_to_file_path(doc_name)

    def _file_path_to_doc_name(self, file_path: str) -> str:
        relpath = super()._file_path_to_doc_name(file_path)
        return self._rel_path_to_module_name(relpath)
        
    def _rel_path_to_module_name(self, rel_path: str) -> str:
        module_name = os.path.splitext(rel_path)[0].replace(os.path.sep, '.')
        return module_name



import pickle
from joblib import Parallel, delayed
import difflib
import typing as t
import re
import os

from ragmatic.utils.refs import RefBaseModel
from pydantic import Field


PARALLELIZATION_THRESHOLD = 1_000


class MapOp:
    
    def __init__(self, config):
        pass
    def map(self, callable, inputs_iterable):
        pass

    def __enter__(self):
        return self
    
    def __exit__(self, exc_type, exc_value, traceback):
        pass


class NoMap(MapOp):
    def map(self, callable, inputs_iterable):
        return [callable(*i) for i in inputs_iterable]


class JoblibParallelMap(MapOp):

    def __init__(self, config):
        self.config = config
        self.parallel = None

    def map(self, callable, inputs_iterable):
        with self.parallel as parallel:
            return parallel(delayed(callable)(*i) for i in inputs_iterable)

    def __enter__(self):
        self.parallel = Parallel(**self.config)
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.parallel = None


class PydictObjStoreConfig(RefBaseModel):
    filepath: t.Optional[str] = Field(default="data.pkl")
    always_parallel: t.Optional[bool] = False
    allowed_query_keys: t.Optional[list[str]] = Field(default_factory=list)
    overwrite: t.Optional[bool] = Field(default=True)


class PydictObjStore:
    name = 'pydict'
    _allowed_query_keys = []
    """
    Provides a list of supported query keys. The keys can be simple or nested.
    The nested keys are separated by a dot. The nested keys can also contain
    a wildcard `{*}` which is used to indicate that the key should be applied
    to all items in a collection.

    Example:
    [
        "name",
        "imports",
        "classes",
        "metrics.instability",
        "metrics.abstractness",
        "metrics.distance_from_main",
        r"classes.{*}.methods",
        r"classes.{*}.methods.{*}.return_type",
        r"classes.{*}.bases",
        r"classes.{*}.metrics.CBO",
        r"classes.{*}.metrics.DIT",
        r"classes.{*}.metrics.Ce",
        r"classes.{*}.metrics.Ca",
        r"classes.{*}.metrics.LCOM4",
        r"classes.{*}.is_abstract",
        r"functions.{*}.metrics.cyclomatic_complexity",
        r"functions.{*}.return_type",
    ]
    """

    _operators = {
        "eq": lambda x, y: x == y,
        "ne": lambda x, y: x != y,
        "gt": lambda x, y: x > y,
        "lt": lambda x, y: x < y,
        "gte": lambda x, y: x >= y,
        "lte": lambda x, y: x <= y,
        "countEq": lambda x, y: len(x) == y,
        "countNe": lambda x, y: len(x) != y,
        "countGt": lambda x, y: len(x) > y,
        "countLt": lambda x, y: len(x) < y,
        "countGte": lambda x, y: len(x) >= y,
        "countLte": lambda x, y: len(x) <= y,
        "in": lambda x, y: y in x,
        "nin": lambda x, y: y not in x,
        "regex": lambda x, y: bool(re.match(y, x)),
    }

    _collection_operators = [
        "countEq",
        "countNe",
        "countGt",
        "countLt",
        "countGte",
        "countLte",
        "in",
        "nin",
    ]

    _default_filepath = "data.pkl"

    def __init__(self, config):
        config = PydictObjStoreConfig(**config)
        self.config = config
        self.overwrite = config.overwrite
        self.always_parallel = config.always_parallel
        self.filepath = os.path.expanduser(config.filepath)
        self._allowed_query_keys = config.allowed_query_keys or self._allowed_query_keys
        self.__data = None

    @property
    def _data(self):
        if self.__data is None:
            self._load_module_data()
        return self.__data
    
    def store_data(self, data: dict[str, t.Any]):
        if self.overwrite:
            self.__data = data
        else:
            self._data.update(data)
        self._write_data(self._data, self.filepath)

    def _write_data(self, data, filepath):
        with open(filepath, "wb") as f:
            pickle.dump(data, f)


    def get_mapper(self):
        if len(self._data) > PARALLELIZATION_THRESHOLD or self.always_parallel:
            return JoblibParallelMap({"n_jobs": -1})
        return NoMap({})

    def _load_module_data(self):
        if not os.path.exists(self.filepath):
            raise FileNotFoundError(
                f"Unable to load data: File {self.filepath} does not exist."
            )
        with open(self.filepath, "rb") as f:
            self.__data = pickle.load(f)

    def query_data(self, query: list[tuple[str, str, t.Any]]):
        query_ops = []
        for line in query:
            self._validate_query_line(line)
            query_ops.append(line)
        first_op = query_ops[0]
        results = self._apply_query_op(first_op)
        if len(query_ops) == 1:
            return results
        for op in query_ops[1:]:
            results = results.intersection(self._apply_query_op(op))
        return results

    def _validate_query_line(self, query_line: tuple[str, str, t.Any]):
        key, op, value = query_line
        self._validate_query_key(key)
        self._validate_operator(op)
    
    def _extract_value_from_key(self, key, module_data: t.Any):
        keyparts = key.split(".")
        if hasattr(module_data, "name"):
            module_name = module_data.name
        elif hasattr(module_data, "__getitem__") and "name" in module_data:
            module_name = module_data["name"]
        else:
            module_name = key
        return module_name, self._extract_value_from_keyparts(keyparts, module_data)

    def _extract_value_from_keyparts(self, keyparts, obj):
        value = obj
        for i, keypart in enumerate(keyparts):
            if keypart == "{*}":
                value = [
                    self._extract_value_from_keyparts(keyparts[i+1:], _v)
                    for _, _v in value.items()
                ]
                break
            elif hasattr(value, keypart):
                value = getattr(value, keypart)
            elif hasattr(value, "__getitem__") and keypart in value:
                value = value[keypart]
            else:
                raise ValueError(f"Key {keypart} not found in object")
        return self._flatten(value)

    def _flatten(self, value: t.Union[t.Any, t.List[t.Any], t.List[t.List[t.Any]]]):
        if isinstance(value, (list, dict, tuple)):
            new_value = []
            for v in value:
                if isinstance(v, (list, dict, tuple)):
                    new_value.extend(self._flatten(v))
                else:
                    new_value.append(v)
            return new_value
        return value
    
    def _validate_query_key(self, key):
        if key not in self._allowed_query_keys:
            suggested_keys = difflib.get_close_matches(key, self._allowed_query_keys)
            if suggested_keys:
                raise ValueError(f"Query key {key} not allowed. Did you mean one of these: {suggested_keys}")
            else:
                raise ValueError(f"Query key {key} not allowed")

    def _validate_operator(self, operator):
        if operator not in self._operators:
            suggested_ops = difflib.get_close_matches(operator, self._operators.keys())
            if suggested_ops:
                raise ValueError(f"Query operator {operator} not allowed. Did you mean one of these: {suggested_ops}")
            else:
                raise ValueError(f"Query operator {operator} not allowed")

    def _apply_query_op(self, query_op: tuple[str, str, t.Any]):
        key, op, value = query_op
        with self.get_mapper() as mapper:
            output_values = dict(mapper.map(self._extract_value_from_key, [(key, m) for m in self._data.values()]))
            output_results = dict(
                mapper.map(
                    self._apply_operator, 
                    [(module_name, op, value, mod_value) for module_name, mod_value in output_values.items()]
                )
            )
        return {k for k,v in output_results.items() if v}
    
    def _apply_operator(self, module_name, operator, value, mod_value):
        
        if isinstance(mod_value, list):
            if operator in self._collection_operators:
                return module_name, self._operators[operator](mod_value, value)
            else:
                if len(mod_value) == 0:
                    return module_name, False
                return module_name, all([self._operators[operator](m, value) for m in mod_value])
        else:
            return module_name, self._operators[operator](mod_value, value)

    def get_data(self, key: str):
        return self._data.get(key)


    


from typing import List, Callable, Optional

from ragmatic.utils.refs import RefBaseModel
from pydantic import Field, ConfigDict
from ..llm_ops.bases import LLMClientBase
from ..llm_ops.client_factory import get_llm_client_class
from ..storage.store_factory import get_store_cls
from ..storage.bases import VectorStore
from ..embeddings.bases import Embedder
from ..embeddings.embedder_factory import get_embedder_cls
from ragmatic.utils import CollectionKeyFormatter
from ragmatic.common_types import TypeAndConfig, StoreConfig
from ..document_sources.bases import DocumentSourceBase

class RagAgentConfig(RefBaseModel):
    llm: TypeAndConfig
    storage: StoreConfig
    encoder: TypeAndConfig
    n_nearest: Optional[int] = 10
    prompt: Optional[str] = Field(default="")
    system_prompt: Optional[str] = Field(default="")
    model_config = ConfigDict(extra= "allow")

class RagAgentBase:

    name: str = ""
    file_filters: List[Callable[[str], bool]] = [(lambda x: True)]
    prompt: str = ""
    system_prompt: str = ""
    q_context_delimiter: str = "\n=========\n"

    def __init__(self, config: RagAgentConfig, document_source: DocumentSourceBase):
        self.config = config
        self.prompt = config.prompt or self.prompt
        self.system_prompt = config.system_prompt or self.system_prompt
        self._document_source = document_source
        self._n = config.n_nearest
        self._llm_client: LLMClientBase = self._initialize_llm_client()
        self._vector_store: VectorStore = self._initialize_vector_store()
        self._embedder: Embedder = self._initialize_embedder()

    def _initialize_llm_client(self) -> LLMClientBase:
        client_class = get_llm_client_class(self.config.llm.type)
        llm_config = self.config.llm.config
        return client_class(llm_config)
    
    def _initialize_vector_store(self) -> VectorStore:
        store_class = get_store_cls(self.config.storage.data_type, self.config.storage.type)
        store_config = self.config.storage.config
        return store_class(store_config)

    def _initialize_embedder(self) -> Embedder:
        embedder_class = get_embedder_cls(self.config.encoder.type)
        embedder_config = self.config.encoder.config
        return embedder_class(embedder_config)
    
    def query(self, query: str):
        encoded_query = self._embedder.encode([query], query=True)[0]
        doc_name_matches = self._vector_store.query_byvector(encoded_query, self._n)
        doc_names = [CollectionKeyFormatter.extract_collection_name(match) for match in doc_name_matches]
        context_docs = self._document_source.get_documents(doc_names)
        message = self.build_user_message(query, context_docs)
        return self._llm_client.send_message(
            message,
            system_prompt=self.system_prompt
        )

    def build_user_message(self, query: str, context_docs: dict[str, str]):
        raise NotImplementedError



from ._types import PresetData
from .local_docs_preset import local_docs_preset
from .pycode_preset import pycode_preset


_presets = {
    "local_docs": local_docs_preset,
    "pycode": pycode_preset
}

def get_preset(name) -> PresetData:
    if name == "default":
        name = "local_docs"
    if name not in _presets:
        raise ValueError(f"Unknown preset: {name}")
    return _presets[name]



from abc import ABC, abstractmethod
import typing as t


class VectorStore(ABC):
    
    name: str = None

    @abstractmethod
    def store_vectors(self, vectors: dict[str, t.Any]):
        pass

    @abstractmethod
    def get_vectors(self, keys: list[str]):
        pass

    @abstractmethod
    def scan_keys(self, match: str):
        pass

    @abstractmethod
    def query(self, query: t.Any):
        pass

    @abstractmethod
    def query_byvector(self, vector: t.Sequence[float], n: int = None):
        pass


class TextDocumentStore(ABC):

    name: str = None

    @abstractmethod
    def store_text_docs(self, text_docs: dict[str, str]):
        pass

    @abstractmethod
    def get_document(self, key: str) -> str:
        pass

    @abstractmethod
    def get_documents(self, keys: list[str]) -> list[str]:
        pass

    @abstractmethod
    def get_all_documents(self):
        pass


class OmniStore(VectorStore, TextDocumentStore):
    
    _vector_store: VectorStore = None
    _text_doc_store: TextDocumentStore = None


    def __init__(self, config):
        raise NotImplementedError

    def store_vectors(self, vectors: dict[str, t.Any]):
        return self._vector_store.store_vectors(vectors)
    
    def get_vectors(self, keys: list[str]):
        return self._vector_store.get_vectors(keys)
    
    def scan_keys(self, match: str):
        return self._vector_store.scan_keys(match)
    
    def query(self, query: t.Any):
        return self._vector_store.query(query)
    
    def query_byvector(self, vector: t.Sequence[float], n: int = None):
        return self._vector_store.query_byvector(vector, n)
    
    def store_text_docs(self, documents: dict[str, str]):
        return self._text_doc_store.store_text_docs(documents)
    
    def get_document(self, key: str):
        return self._text_doc_store.get_document(key)
    
    def get_documents(self, keys: list[str]):
        return self._text_doc_store.get_documents(keys)
    
    def get_all_documents(self):
        return self._text_doc_store.get_all_documents()

    def __getattr__(self, name):
        if hasattr(self._vector_store, name):
            return getattr(self._vector_store, name)
        if hasattr(self._text_doc_store, name):
            return getattr(self._text_doc_store, name)
        raise AttributeError(f"Attribute {name} not found in {type(self).__name__}.")



import typing as t
from .bases import RagAgentBase
from ..utils import import_object

_rag_agents =  {
    "python_code": "ragmatic.rag.pycode_rag_agent.PyCodeRagAgent",
    "generic": "ragmatic.rag.generic.GenericRagAgent",
}


def get_rag_agent_class(rag_agent_name:str) -> t.Type[RagAgentBase]:
    if rag_agent_name not in _rag_agents:
        try:
            return import_object(rag_agent_name)
        except Exception:
            raise ValueError(f"No RagAgent named {rag_agent_name!r} was found.")
    return import_object(_rag_agents[rag_agent_name])



from typing import List
import re
import os

from .bases import SummarizerBase


class PyCodeSummarizer(SummarizerBase):

    name = "python_code"
    file_filters: List = [(lambda x: x.endswith('.py'))]

    _system_prompt = (
        "You are an assistant who is an expert in Python programming who pays "
        "careful attention to details of code structure, but is also able to "
        "provide high-level summaries of code and its functionality."
    )
    _document_prompt = (
        "Please review the following Python code and provide several summaries. "
        "The first summary should be a high-level overview of the code's functionality. "
        "Then, create a summary for each function or class present in the code. "
        "Finally, provide a summary for any other important components of the code."
        "Identify each summary by a set of <summary></summary> tags."
    )

    def summarize_document(self, doc: str) -> list[str]:
        message = self._build_message(doc)
        response = self._llm_client.send_message(
            message,
            self._system_prompt,
            role="user"
        )
        return re.findall(r"<summary>(.*?)</summary>", response, re.DOTALL)
    
    def _file_path_to_doc_name(self, file_path: str) -> str:
        rel_path = super()._file_path_to_doc_name(file_path)
        return self._file_path_to_module_name(rel_path)

    def _file_path_to_module_name(self, rel_path: str) -> str:
        # strip dots from the beginning of the path
        rel_path = re.sub(r"^\.*", "", rel_path)
        if rel_path.startswith(os.path.sep):
            rel_path = rel_path[1:]
        module_name = os.path.splitext(rel_path)[0].replace(os.path.sep, '.')
        return module_name



from .bases import DocumentSourceBase
from ..storage.store_factory import get_store_cls
from ..storage.bases import TextDocumentStore
from ..common_types import StoreConfig



class TextStoreDocumentSource(DocumentSourceBase):
    
    name = "storage"

    def __init__(self, config: StoreConfig):
        if isinstance(config, dict):
            config = StoreConfig(**config)
        super().__init__(config)
        self._text_doc_store: TextDocumentStore = self._initialize_text_doc_store()
    
    def _initialize_text_doc_store(self):
        store_cls = get_store_cls(self.config.data_type, self.config.type)
        store_config = self.config.config
        return store_cls(store_config)

    def get_documents(self, document_names: list[str] = None) -> dict[str, str]:
        if document_names:
            return dict(
                zip(
                    document_names,
                    self._text_doc_store.get_documents(document_names)
                )
            )
        return self._text_doc_store.get_all_documents()



from string import Template
import typing as t
import json

from ragmatic.utils.refs import (
    ref_dumper_default,
    RefDecoder,
)
from .._types import MasterConfig


class PresetData:

    def __init__(self,
                 components,
                 pipelines,
                 rag_query_command,
                 variable_defaults: dict = None
                 ):
        self.components = components
        self.pipelines = pipelines
        self.rag_query_command = rag_query_command
        self.variable_defaults: dict = variable_defaults or {}

    def get_config(self, **vars):
        return MasterConfig(
            project_name="",
            components=self.get_component_config(**vars),
            pipelines=self.get_pipelines_config(**vars),
            rag_query_command=self.get_rag_query_command_config(**vars)
        )

    def get_component_config(self, **vars):
        return self._apply_variables(self.components, **vars)

    def get_pipelines_config(self, **vars):
        return self._apply_variables(self.pipelines, **vars)
    
    def get_rag_query_command_config(self, **vars):
        return self._apply_variables(self.rag_query_command, **vars)

    def _apply_variables(self, config: t.Dict, **vars):
        for v in vars:
            if v not in self.variable_defaults:
                raise ValueError(f"Unknown variable: {v}")
        variables = self.variable_defaults.copy()
        variables.update(vars)
        json_config = json.dumps(config, default=ref_dumper_default)
        json_config = Template(json_config).substitute(variables)
        return json.loads(json_config, cls=RefDecoder)
    


import typing as t
import pickle
import os

from ragmatic.utils.refs import RefBaseModel
from pydantic import Field

from ..bases import TextDocumentStore


class PydictTextDocumentStoreConfig(RefBaseModel):
    overwrite: t.Optional[bool] = Field(default=False)
    filepath: t.Optional[str] = Field(default=None)
    allow_init: t.Optional[bool] = Field(default=True)


class PydictTextDocumentStore(TextDocumentStore):
    
    name = 'pydict'
    _default_filepath = 'text_documents.pkl'


    def __init__(self, config: PydictTextDocumentStoreConfig):
        config = PydictTextDocumentStoreConfig(**config)
        self.config = config
        self.overwrite = config.overwrite
        self.allow_init = config.allow_init
        self.filepath = self.config.filepath or self._default_filepath
        self.__data: dict[str, str] = {}

    @property
    def _data(self):
        if not self.__data:
            self._load_documents()
        return self.__data

    def store_text_docs(self, text_docs: dict[str, str]):
        if self.overwrite:
            self.__data = text_docs
        else:
            self._data.update(text_docs)
        self._write_documents(self._data)

    def _write_documents(self, data):
        with open(self.filepath, "wb") as f:
            pickle.dump(data, f)

    def _load_documents(self):
        if not os.path.exists(self.filepath):
            if self.allow_init:
                return
            raise FileNotFoundError(
                f"Summaries not loaded: File {self.filepath} does not exist."
            )   
        with open(self.filepath, "rb") as f:
            self.__data = pickle.load(f)

    def get_document(self, key: str):
        return self._data.get(key)

    def get_documents(self, keys: list[str]) -> list[str]:
        docs = []
        for key in keys:
            if key not in self._data:
                raise KeyError(f"Document with key {key} not found in storage.")
            docs.append(self._data.get(key))
        return docs

    def get_all_documents(self) -> dict[str, str]:
        return self._data



from logging import getLogger
from pathlib import Path

from deepmerge import always_merger, Merger

from ragmatic.utils import ragmatic_load_yaml
from ragmatic.utils.refs import resolve_references
from ._types import MasterConfig
from ..configuration.presets.preset_factory import get_preset, PresetData

logger = getLogger(__name__)


def load_configdict(configpath: Path = None) -> dict:
    with open(str(configpath)) as f:
        config = ragmatic_load_yaml(f)
    return config


def load_config(configpath: Path = None) -> MasterConfig:
    config = load_configdict(configpath)
    return MasterConfig(**config)


def get_preset_config(preset_name, **vars) -> MasterConfig:
    preset = get_preset(preset_name)
    return preset.get_config(**vars)


def merge_defaults(config: MasterConfig,
                   preset_data: PresetData = None,
                   **vars
                   ) -> MasterConfig:
    config_d = config.model_dump()
    component_config = config_d.get("components", {})
    pipelines_config = config_d.get("pipelines", {})
    rag_query_command = config_d.get("rag_query_command", {})
    
    component_config = always_merger.merge(
        preset_data.get_component_config(**vars),
        component_config
    )
    config_d.update({"components": component_config})

    pipelines_merger = Merger(
        [
            (dict, "merge"),
            (list, "append")
        ],
        ["override"],
        ["override"]
    )
    pipelines_config = pipelines_merger.merge(
        preset_data.get_pipelines_config(**vars),
        pipelines_config
    )
    rag_query_command = always_merger.merge(
        preset_data.get_rag_query_command_config(**vars),
        rag_query_command
    )
    return MasterConfig(
        project_name=config.project_name,
        components=component_config,
        pipelines=pipelines_config,
        rag_query_command=rag_query_command
    )

def resolve_config_references(config: MasterConfig) -> MasterConfig:
    config_dict = config.model_dump()
    resolved_config = resolve_references(config_dict)
    return MasterConfig(**resolved_config)
